@@ -17,15 +17,8 @@ impl<'info> Validate<'info> for NewPool<'info> {
     fn validate(&self) -> Result<()> {
         // replica_mint checks are now redundant
         // since it is now an associated mint
-        assert_keys_eq!(
-            self.replica_mint.mint_authority.unwrap(),
-            self.pool,
-            "replica_mint.mint_authority"
-        );
-        invariant!(
-            self.replica_mint.freeze_authority.is_none(),
-            "cannot have freeze authority"
-        );
+        assert_keys_eq!(self.replica_mint.mint_authority.unwrap(), self.pool);
+        invariant!(self.replica_mint.freeze_authority.is_none());
         invariant!(
             self.replica_mint.decimals == self.primary_mint.decimals,
             "decimals mismatch"
@@ -67,34 +60,32 @@ impl<'info> Validate<'info> for InitMiner<'info> {
 
 impl<'info> Validate<'info> for WithdrawTokens<'info> {
     fn validate(&self) -> Result<()> {
+        assert_keys_eq!(self.owner, self.mm.owner, Unauthorized);
+        assert_keys_eq!(self.pool, self.mm.pool);
+
         let withdraw_mint = self.mm_token_account.mint;
 
-        assert_keys_eq!(self.withdraw_mint, withdraw_mint, "withdraw_mint");
+        assert_keys_eq!(self.withdraw_mint, withdraw_mint);
         // cannot withdraw a replica mint.
         assert_keys_neq!(
-            self.withdraw_mint.key(),
+            self.withdraw_mint,
             self.pool.replica_mint,
             CannotWithdrawReplicaMint
         );
 
+        assert_keys_neq!(self.mm_token_account, self.token_destination);
+
         if withdraw_mint == self.pool.primary_mint {
             // should be no replica balance if withdrawing primary
             invariant!(self.mm.replica_balance == 0, OutstandingReplicaTokens);
         }
 
-        assert_keys_eq!(self.owner, self.mm.owner);
-        assert_keys_eq!(self.pool, self.mm.pool);
-
         assert_keys_eq!(self.mm_token_account.mint, withdraw_mint);
         assert_keys_eq!(self.mm_token_account.owner, self.mm);
         invariant!(self.mm_token_account.delegate.is_none());
         invariant!(self.mm_token_account.close_authority.is_none());
 
-        assert_keys_eq!(
-            self.token_destination.mint,
-            withdraw_mint,
-            "token_destination.mint"
-        );
+        assert_keys_eq!(self.token_destination.mint, withdraw_mint);
 
         Ok(())
     }
@@ -104,6 +95,9 @@ impl<'info> Validate<'info> for ClaimRewards<'info> {
     fn validate(&self) -> Result<()> {
         self.stake.validate()?;
 
+        let staked_mint_key = self.stake.quarry.token_mint_key;
+
+        assert_keys_eq!(self.mint_wrapper, self.stake.rewarder.mint_wrapper);
         assert_keys_eq!(self.minter.mint_wrapper, self.mint_wrapper);
         assert_keys_eq!(self.minter.minter_authority, self.stake.rewarder);
         assert_keys_eq!(self.rewards_token_mint, self.mint_wrapper.token_mint);
@@ -111,11 +105,12 @@ impl<'info> Validate<'info> for ClaimRewards<'info> {
         assert_keys_eq!(self.rewards_token_account.mint, self.rewards_token_mint);
         assert_keys_eq!(self.rewards_token_account.owner, self.stake.mm);
 
-        assert_keys_eq!(self.claim_fee_token_account.mint, self.rewards_token_mint);
         assert_keys_eq!(
-            self.stake_token_account.mint,
-            self.stake.quarry.token_mint_key
+            self.claim_fee_token_account,
+            self.stake.rewarder.claim_fee_token_account
         );
+        assert_keys_eq!(self.claim_fee_token_account.mint, self.rewards_token_mint);
+        assert_keys_eq!(self.stake_token_account.mint, staked_mint_key);
 
         Ok(())
     }
@@ -129,20 +124,15 @@ impl<'info> Validate<'info> for QuarryStakePrimary<'info> {
     fn validate(&self) -> Result<()> {
         self.stake.validate()?;
 
+        assert_keys_eq!(self.mm_owner, self.stake.mm.owner);
+
         // For primary staking:
         // - quarry is a quarry, staking the `primary_mint`
 
-        assert_keys_eq!(self.mm_owner, self.stake.mm.owner, "mm_owner");
-        assert_keys_eq!(
-            self.stake.quarry.token_mint_key,
-            self.stake.pool.primary_mint,
-            "stake.quarry.token_mint_key"
-        );
+        let primary_mint = self.stake.quarry.token_mint_key;
+        assert_keys_eq!(primary_mint, self.stake.pool.primary_mint);
+        assert_keys_eq!(primary_mint, self.mm_primary_token_account.mint);
 
-        assert_keys_eq!(
-            self.mm_primary_token_account.mint,
-            self.stake.pool.primary_mint
-        );
         assert_keys_eq!(self.mm_primary_token_account.owner, self.stake.mm);
         invariant!(self.mm_primary_token_account.delegate.is_none());
         invariant!(self.mm_primary_token_account.close_authority.is_none());
@@ -158,18 +148,11 @@ impl<'info> Validate<'info> for QuarryStakeReplica<'info> {
         // For replica staking:
         // - rewarder is any rewarder
         // - quarry is a quarry staking the `replica_mint`
-        assert_keys_eq!(self.mm_owner, self.stake.mm.owner, "mm_owner");
+        let replica_mint_key = self.replica_mint.key();
+        assert_keys_eq!(self.mm_owner, self.stake.mm.owner);
 
-        assert_keys_eq!(
-            self.stake.pool.replica_mint,
-            self.replica_mint,
-            "stake.pool.replica_mint"
-        );
-        assert_keys_eq!(
-            self.stake.quarry.token_mint_key,
-            self.replica_mint,
-            "stake.quarry.token_mint_key"
-        );
+        assert_keys_eq!(self.stake.pool.replica_mint, replica_mint_key);
+        assert_keys_eq!(self.stake.quarry.token_mint_key, replica_mint_key);
 
         assert_keys_eq!(self.replica_mint_token_account.mint, self.replica_mint);
         assert_keys_eq!(self.replica_mint_token_account.owner, self.stake.mm);
@@ -182,14 +165,21 @@ impl<'info> Validate<'info> for QuarryStakeReplica<'info> {
 
 impl<'info> Validate<'info> for QuarryStake<'info> {
     fn validate(&self) -> Result<()> {
-        assert_keys_eq!(self.mm.pool, self.pool, "mm.pool");
+        // this links merge miner validations with quarry validations.
+        let quarry_mint = self.quarry.token_mint_key;
+        invariant!(self.pool.primary_mint == quarry_mint || self.pool.replica_mint == quarry_mint);
 
-        assert_keys_eq!(self.rewarder, self.quarry.rewarder_key);
+        // merge miner validations
+        assert_keys_eq!(self.mm, self.miner.authority);
+        assert_keys_eq!(self.pool, self.mm.pool);
+
+        // Quarry validations
         assert_keys_eq!(self.quarry, self.miner.quarry_key);
-        assert_keys_eq!(self.miner.authority, self.mm, "miner.authority");
+        assert_keys_eq!(self.rewarder, self.quarry.rewarder_key);
 
         assert_keys_eq!(self.miner_vault, self.miner.token_vault_key);
         assert_keys_eq!(self.miner_vault.owner, self.miner);
+        assert_keys_eq!(self.miner_vault.mint, quarry_mint);
         invariant!(self.miner_vault.delegate.is_none());
         invariant!(self.miner_vault.close_authority.is_none());
 
